/***************************************************************************
***
***    Copyright 2005  Hon Hai Precision Ind. Co. Ltd.
***    All Rights Reserved.
***    No portions of this material shall be reproduced in any form without
***    the written permission of Hon Hai Precision Ind. Co. Ltd.
***
***    All information contained in this document is Hon Hai Precision Ind.
***    Co. Ltd. company private, proprietary, and trade secret property and
***    are protected by international intellectual property laws and treaties.
***
****************************************************************************
***
***    FILE NAME :
***      cmd_port.c
***
***    DESCRIPTION :
***      for switch port related setting get/set
***
***    HISTORY :
***       - 2014/1/10 16:36:23, Joyce Peng
***             File Creation
***
***************************************************************************/

/*==========================================================================
 *                                                                          
 *      Library Inclusion Segment
 *                                                                          
 *==========================================================================
 */
#include "system.h"
#include "soc/soc.h"
#include "soc/bcm5340x.h"
#include "utils/net.h"
#include "utils/ports.h"
#include "boardapi/base.h"
#include "boardapi/port.h"


#include "appl/lacp/bonding.h"
#include "appl/lacp/bond_3ad.h"
//#include "appl/rstp/rstp.h" 

#ifdef CFG_SWITCH_LAG_INCLUDED

/*==========================================================================
 *
 *      Constant
 *
 *==========================================================================
 */

/*==========================================================================
 *
 *      Structrue segment
 *
 *==========================================================================
 */

/*==========================================================================
 *
 *      Static/Global Variable segment
 *
 *==========================================================================
 */
static struct lacp_bond g_system_bond; 
static BOOL g_lacp_enable            = FALSE;
static BOOL g_lacp_dbg_print         = FALSE;
static struct bonding *g_pSysBonds   = NULL;
static struct slave *g_pSysSlaves    = NULL;
static struct net_device *g_pSysDevs = NULL;
static uint8 g_u8WalkThroughLagID    = 0;

/*==========================================================================
 *                                                                          
 *      Function Definition Segment
 *                                                                          
 *==========================================================================
 */
/* bonding functions/UM hook functions */
static void bond_attach_slave(struct bonding *bond, struct slave *new_slave);
static void bond_detach_slave(struct bonding *bond, struct slave *slave);
static sys_error_t bond_update_speed_duplex(uint16 uport, struct slave *pSlave);
static BOOL allocateLacpMemoryPool(void);
static BOOL freeLacpMemoryPool(void);
static BOOL locate_lacp_bpdu_pkt(uint8 *eth_pkt, struct lacpdu_header *placpdu_frame, struct sk_buff *skb);
static void show_lacpdu(struct sk_buff *skb);
#if 0 /* Ccc: */
static void show_port_channel_info(uint8 u8LagID, struct port_channel_info *pChannelInfo);
#endif
static void lacp_update_dev_flag(uint8 u8LagID);
static sys_rx_t lacp_rx(sys_pkt_t *pkt, void *cookie) REENTRANT;
STATICCBK void lacp_tx_async_cbk(sys_pkt_t *pkt, sys_error_t status) REENTRANT;


/*==========================================================================
 *
 *      Local Function Body segment
 *
 *==========================================================================
 */

/*--------------------------- slave list handling ---------------------------*/

/*
 * This function attaches the slave to the end of list.
 *
 * bond->lock held for writing by caller.
 */
static void bond_attach_slave(struct bonding *bond, struct slave *new_slave)
{
	if (bond->first_slave == NULL) { /* attaching the first slave */
		new_slave->next = new_slave;
		new_slave->prev = new_slave;
		bond->first_slave = new_slave;
	} else {
		new_slave->next = bond->first_slave;
		new_slave->prev = bond->first_slave->prev;
		new_slave->next->prev = new_slave;
		new_slave->prev->next = new_slave;
	}

	bond->slave_cnt++;
}

/*
 * This function detaches the slave from the list.
 * WARNING: no check is made to verify if the slave effectively
 * belongs to <bond>.
 * Nothing is freed on return, structures are just unchained.
 * If any slave pointer in bond was pointing to <slave>,
 * it should be changed by the calling function.
 *
 * bond->lock held for writing by caller.
 */
static void bond_detach_slave(struct bonding *bond, struct slave *slave)
{
	if (slave->next)
		slave->next->prev = slave->prev;

	if (slave->prev)
		slave->prev->next = slave->next;

	if (bond->first_slave == slave) { /* slave is the first slave */
		if (bond->slave_cnt > 1) { /* there are more slave */
			bond->first_slave = slave->next;
		} else {
			bond->first_slave = NULL; /* slave was the last one */
		}
	}

	slave->next = NULL;
	slave->prev = NULL;
	bond->slave_cnt--;
}

static sys_error_t bond_update_speed_duplex(uint16 uport, struct slave *pSlave)
{
    BOOL buLink;
    sys_error_t rv = SYS_OK;
    uint8 speedMode, duplexMode;

    if (NULL==pSlave)
        return SYS_ERR_PARAMETER;

    pSlave->speed = 0;
    pSlave->link = BOND_LINK_DOWN;
    pSlave->duplex = DUPLEX_HALF;

    rv = board_port_link_status_get(uport, &buLink);
    if ((SYS_OK==rv) && (TRUE==buLink)) {
        pSlave->link = BOND_LINK_UP;
        bond_set_slave_active_flags(pSlave);
        pSlave->dev->flags |= IFF_UP;
    }
    else {
        bond_set_slave_inactive_flags(pSlave);
        pSlave->dev->flags &= ~IFF_UP;
    }

    rv = board_port_speed_duplex_get(uport, &speedMode, &duplexMode);
    /*
        typedef enum {
            PS_FORCE_SPEED_10M,
            PS_FORCE_SPEED_100M,
            PS_FORCE_SPEED_1000M,
            PS_FORCE_SPEED_2500M,
            PS_FORCE_SPEED_5000M,
            PS_FORCE_SPEED_10000M
        } portsetting_forcespeed_t;
    */
    if (SYS_OK==rv) {
        switch (speedMode)
        {
            case PS_FORCE_SPEED_10M:
                pSlave->speed = SPEED_10;
                break;
            case PS_FORCE_SPEED_100M:
                pSlave->speed = SPEED_100;
                break;
            case PS_FORCE_SPEED_1000M:
                pSlave->speed = SPEED_1000;
                break;
            case PS_FORCE_SPEED_10000M:
                pSlave->speed = SPEED_10000;
                break;
            default:
                pSlave->speed = 0;
                break;
        }
        
        switch (duplexMode)
        {
            case 1:
                pSlave->duplex = DUPLEX_FULL;
                break;
                
            case 0:                
            default:
                pSlave->duplex = DUPLEX_HALF;
                break;
        }
    }

    pr_debug("port %d status: pSlave->link=%d, pSlave->speed=%d, pSlave->duplex=%d\n", pSlave->dev->uport, pSlave->link, pSlave->speed, pSlave->duplex);
    
    return rv;
} 

static BOOL allocateLacpMemoryPool(void)
{
    BOOL blResult = TRUE;
    uint8 u8MaxPorts = board_uport_count();
    sys_error_t nLacpErr = SYS_OK;
    int nMemSize;

    if (FALSE==g_lacp_enable) {

        /* Allocate the maximum memory for bond using */
        nMemSize = sizeof(struct bonding) * BOARD_MAX_NUM_OF_LAG;
        g_pSysBonds = malloc(nMemSize);
        if (NULL==g_pSysBonds) {
            LACP_ERR(("Out of memory to create channel group with size (%d)!\n", nMemSize));
            nLacpErr = SYS_ERR_OUT_OF_RESOURCE;
        }
        memset(g_pSysBonds, (char)0, nMemSize);
        
        /* Allocate the maximum memory for slave using */
        nMemSize = sizeof(struct slave) * u8MaxPorts;
        g_pSysSlaves = malloc(nMemSize);
        if (NULL==g_pSysSlaves) {
            LACP_ERR(("Out of memory to create slave for port with size (%d)!\n", nMemSize));
            nLacpErr = SYS_ERR_OUT_OF_RESOURCE;
        }
        memset(g_pSysSlaves, (char)0, nMemSize);
        
        /* Allocate the maximum memory for dev using */
        nMemSize = sizeof(struct net_device) * u8MaxPorts;
        g_pSysDevs = malloc(nMemSize);
        if (NULL==g_pSysDevs) {
            LACP_ERR(("Out of memory to create dev for port with size (%d)!\n", nMemSize));
            nLacpErr = SYS_ERR_OUT_OF_RESOURCE;
        }
        memset(g_pSysDevs, (char)0, nMemSize);
         
        if (SYS_OK!=nLacpErr) {
            if (NULL!=g_pSysBonds)
                free(g_pSysBonds);
            if (NULL!=g_pSysSlaves)
                free(g_pSysSlaves);
            if (NULL!=g_pSysDevs)
                free(g_pSysDevs);
            LACP_ERR(("out of memory and disable LACP !\n"));
        }
    }
    
    return blResult;
}

static BOOL freeLacpMemoryPool(void)
{
    if (NULL!=g_pSysBonds)
        free(g_pSysBonds);
    if (NULL!=g_pSysSlaves)
        free(g_pSysSlaves);
    if (NULL!=g_pSysDevs)
        free(g_pSysDevs);

    g_pSysBonds  = NULL;
    g_pSysSlaves = NULL;
    g_pSysDevs   = NULL;
    
    return TRUE;
}

static BOOL locate_lacp_bpdu_pkt(uint8 *eth_pkt, struct lacpdu_header *placpdu_frame, struct sk_buff *skb)
{
    /* omit the tag if need */
	if (ENET_TAGGED((enet_hdr_t *)eth_pkt)) {
        memcpy(&eth_pkt[ETH_ALEN*2], &eth_pkt[ETH_ALEN*2 + VLAN_TAG_LEN]
                , sizeof(unsigned short)+sizeof(struct lacpdu));
    }
    
    memcpy((uint8 *)placpdu_frame, eth_pkt, sizeof(struct lacpdu_header));

    if (AD_TYPE_LACPDU==placpdu_frame->lacpdu.subtype) {

        skb->pdu_type = AD_TYPE_LACPDU;
    }
    else if (AD_TYPE_MARKER==placpdu_frame->lacpdu.subtype) {
        /* No need to convert fields to Little Endian since we don't use the marker's fields. */

        skb->pdu_type = AD_TYPE_MARKER;
    }
    else
        return FALSE;

    if (ntohs(ETH_P_SLOW) != placpdu_frame->hdr.h_proto)
        return FALSE;

    skb->protocol = AD_PKT_TYPE_LACPDU;
    skb->len = sizeof(struct lacpdu_header);        /* lacpdu frame size is same with maker pdu */
    skb->data = (void *)placpdu_frame;

    return TRUE;
}

static void show_lacpdu(struct sk_buff *skb)
{
    int nIdx;
    struct lacpdu_header *placpdu_frame = (struct lacpdu_header *)skb->data;
    struct bond_marker_header *pmarker_frame = NULL;
/*
    pr_debug("skb->protocol=0x%x\n", skb->protocol);
    pr_debug("skb->pdu_type=%d\n", skb->pdu_type);
    pr_debug("skb->len=%d\n", skb->len);
*/    
    if (AD_TYPE_LACPDU==placpdu_frame->lacpdu.subtype) {
#if 1
        pr_debug("placpdu_frame->lacpdu.actor_port=0x%x, actor_state=0x%x\n", ntohs(placpdu_frame->lacpdu.actor_port), placpdu_frame->lacpdu.actor_state);
        pr_debug("placpdu_frame->lacpdu.partner_port=0x%x, partner_state=0x%x\n", ntohs(placpdu_frame->lacpdu.partner_port), placpdu_frame->lacpdu.partner_state);
#else
        for (nIdx=0; nIdx<ETH_ALEN; nIdx++)
            pr_debug("placpdu_frame->hdr.h_dest[%d]=0x%x\n", nIdx, placpdu_frame->hdr.h_dest[nIdx]);
        for (nIdx=0; nIdx<ETH_ALEN; nIdx++)
            pr_debug("placpdu_frame->hdr.h_source[%d]=0x%x\n", nIdx, placpdu_frame->hdr.h_source[nIdx]);
        
        pr_debug("placpdu_frame->hdr.h_proto=0x%x\n", ntohs(placpdu_frame->hdr.h_proto));
        pr_debug("placpdu_frame->lacpdu.subtype=0x%x\n", placpdu_frame->lacpdu.subtype);
        pr_debug("placpdu_frame->lacpdu.version_number=0x%x\n", placpdu_frame->lacpdu.version_number);
        pr_debug("placpdu_frame->lacpdu.tlv_type_actor_info=0x%x\n", placpdu_frame->lacpdu.tlv_type_actor_info);

        pr_debug("placpdu_frame->lacpdu.actor_information_length=0x%x\n", placpdu_frame->lacpdu.actor_information_length);
        pr_debug("placpdu_frame->lacpdu.actor_system_priority=0x%x\n", ntohs(placpdu_frame->lacpdu.actor_system_priority));

        for (nIdx=0; nIdx<ETH_ALEN; nIdx++)
            pr_debug("placpdu_frame->lacpdu.actor_system.mac_addr_value[%d]=0x%x\n", nIdx, placpdu_frame->lacpdu.actor_system.mac_addr_value[nIdx]);

        pr_debug("placpdu_frame->lacpdu.actor_key=0x%x\n", ntohs(placpdu_frame->lacpdu.actor_key));
        pr_debug("placpdu_frame->lacpdu.actor_port_priority=0x%x\n", ntohs(placpdu_frame->lacpdu.actor_port_priority));
        pr_debug("placpdu_frame->lacpdu.actor_port=0x%x\n", ntohs(placpdu_frame->lacpdu.actor_port));
        pr_debug("placpdu_frame->lacpdu.actor_state=0x%x\n", placpdu_frame->lacpdu.actor_state);

        pr_debug("placpdu_frame->lacpdu.tlv_type_partner_info=0x%x\n", placpdu_frame->lacpdu.tlv_type_partner_info);
        pr_debug("placpdu_frame->lacpdu.partner_information_length=0x%x\n", placpdu_frame->lacpdu.partner_information_length);
        pr_debug("placpdu_frame->lacpdu.partner_system_priority=0x%x\n", ntohs(placpdu_frame->lacpdu.partner_system_priority));

        for (nIdx=0; nIdx<ETH_ALEN; nIdx++)
            pr_debug("placpdu_frame->lacpdu.partner_system.mac_addr_value[%d]=0x%x\n", nIdx, placpdu_frame->lacpdu.partner_system.mac_addr_value[nIdx]);

        pr_debug("placpdu_frame->lacpdu.partner_key=0x%x\n", ntohs(placpdu_frame->lacpdu.partner_key));
        pr_debug("placpdu_frame->lacpdu.partner_port_priority=0x%x\n", ntohs(placpdu_frame->lacpdu.partner_port_priority));
        pr_debug("placpdu_frame->lacpdu.partner_port=0x%x\n", ntohs(placpdu_frame->lacpdu.partner_port));
        pr_debug("placpdu_frame->lacpdu.partner_state=0x%x\n", placpdu_frame->lacpdu.partner_state);

        pr_debug("placpdu_frame->lacpdu.tlv_type_collector_info=0x%x\n", placpdu_frame->lacpdu.tlv_type_collector_info);
        pr_debug("placpdu_frame->lacpdu.collector_information_length=0x%x\n", placpdu_frame->lacpdu.collector_information_length);
        pr_debug("placpdu_frame->lacpdu.collector_max_delay=0x%x\n", placpdu_frame->lacpdu.collector_max_delay);

        pr_debug("placpdu_frame->lacpdu.tlv_type_terminator=0x%x\n", placpdu_frame->lacpdu.tlv_type_terminator);
        pr_debug("placpdu_frame->lacpdu.terminator_length=0x%x\n", placpdu_frame->lacpdu.terminator_length);
#endif        
    }
    else if (AD_TYPE_MARKER==placpdu_frame->lacpdu.subtype) {
        
        pmarker_frame = (struct bond_marker_header *)skb->data;

        for (nIdx=0; nIdx<ETH_ALEN; nIdx++)
            pr_debug("pmarker_frame->hdr.h_dest[%d]=0x%x\n", nIdx, pmarker_frame->hdr.h_dest[nIdx]);
        for (nIdx=0; nIdx<ETH_ALEN; nIdx++)
            pr_debug("pmarker_frame->hdr.h_source[%d]=0x%x\n", nIdx, pmarker_frame->hdr.h_source[nIdx]);
        
        pr_debug("pmarker_frame->hdr.h_proto=0x%x\n", pmarker_frame->hdr.h_proto);
        pr_debug("pmarker_frame->marker.subtype=0x%x\n", pmarker_frame->marker.subtype);
        
    }
    else {
        pr_debug("Unknown BPDU subtype (%d) !\n", placpdu_frame->lacpdu.subtype);
    }
}

#if 0 /* Ccc: */
static void show_port_channel_info(uint8 u8LagID, struct port_channel_info *pChannelInfo)
{
    uint8 u8PortCnt;
    struct lacp_port_info *pPortInfo = NULL;

    if (NULL==pChannelInfo)
        return;

    pr_debug("---------------------------------------------------------------\n");
    pr_debug("show_port_channel_info() on LAG ID = %d\n", u8LagID);    
    for(u8PortCnt=0; u8PortCnt < pChannelInfo->u8Ports; u8PortCnt++) {
        pPortInfo = &pChannelInfo->pLacp_port_info[u8PortCnt];

        pr_debug("uport=%d\n", pPortInfo->uport+1);

        pr_debug("actor system mac_addr_value=%x %x %x %x %x %x\n"
                    , pPortInfo->actor_oper.system.mac_addr_value[0]
                    , pPortInfo->actor_oper.system.mac_addr_value[1]
                    , pPortInfo->actor_oper.system.mac_addr_value[2]
                    , pPortInfo->actor_oper.system.mac_addr_value[3]
                    , pPortInfo->actor_oper.system.mac_addr_value[4]
                    , pPortInfo->actor_oper.system.mac_addr_value[5]);
        pr_debug("actor system_priority=%d\n", pPortInfo->actor_oper.system_priority);
        pr_debug("actor port_number=%d\n", pPortInfo->actor_oper.port_number);
        pr_debug("actor port_priority=%d\n", pPortInfo->actor_oper.port_priority);
        pr_debug("actor port_state=0x%x\n", pPortInfo->actor_oper.port_state);
        pr_debug("actor port_key=%d\n", pPortInfo->actor_oper.key);

        pr_debug("partner system mac_addr_value=%x %x %x %x %x %x\n"
                    , pPortInfo->partner_oper.system.mac_addr_value[0]
                    , pPortInfo->partner_oper.system.mac_addr_value[1]
                    , pPortInfo->partner_oper.system.mac_addr_value[2]
                    , pPortInfo->partner_oper.system.mac_addr_value[3]
                    , pPortInfo->partner_oper.system.mac_addr_value[4]
                    , pPortInfo->partner_oper.system.mac_addr_value[5]);
        pr_debug("partner system_priority=%d\n", pPortInfo->partner_oper.system_priority);
        pr_debug("partner port_number=%d\n", pPortInfo->partner_oper.port_number);
        pr_debug("partner port_priority=%d\n", pPortInfo->partner_oper.port_priority);
        pr_debug("partner port_state=0x%x\n", pPortInfo->partner_oper.port_state);
        pr_debug("partner port key=%d\n", pPortInfo->partner_oper.key);

        pr_debug("---------------------------------------------------------------\n");
    }
}
#endif

/* Update the IFF_SLAVE/IFF_MASTER flag */
static void lacp_update_dev_flag(uint8 u8LagID)
{
    int nCnt;
    struct bonding *pBond = NULL;
    struct slave *pSlave = NULL;
    struct net_device *pMasterDev = NULL;

    pBond = g_system_bond.lacp_channel[u8LagID].pBond;
    if (NULL != pBond) {
        for (nCnt=0, pSlave=pBond->first_slave; nCnt<pBond->slave_cnt; nCnt++, pSlave = pSlave->next) {

            /* first slave = master slave */
            if (0==nCnt) {
                pMasterDev = pSlave->dev;
                pBond->dev = pMasterDev;
                pSlave->dev->flags |= IFF_MASTER;
                pSlave->dev->flags &= ~IFF_SLAVE;
                
                pSlave->dev->master = pMasterDev;
            }
            else {
                pSlave->dev->flags |= IFF_SLAVE;
                pSlave->dev->flags &= ~IFF_MASTER;
                pSlave->dev->master = pMasterDev;
            }
        }
    }    
}

static sys_rx_t lacp_rx(sys_pkt_t *pkt, void *cookie) REENTRANT
{
    UNREFERENCED_PARAMETER(cookie);
    struct sk_buff skb;
    struct bonding *pBond = NULL;
    struct slave *pSlave = NULL;
    struct lacpdu_header lacpdu_frame;
    u8 lacpdu_mcast_addr[ETH_ALEN] = MULTICAST_LACPDU_ADDR;

    if ( (FALSE==g_lacp_enable)
         || (NULL==pkt)
         || (memcmp(pkt->pkt_data, lacpdu_mcast_addr, sizeof(lacpdu_mcast_addr))) ) {
         return SYS_RX_NOT_HANDLED;
    }

    /* convert the lacp packet */
    memset((char *)&skb, (char)0, sizeof(skb));
    if (FALSE==locate_lacp_bpdu_pkt(pkt->pkt_data, &lacpdu_frame, &skb))
        return SYS_RX_NOT_HANDLED;

    /* get the bond/slave pointer by lport */
    if (TRUE==lacp_get_bond_slave_by_lport(pkt->rx_src_uport, NULL, &pBond, &pSlave)) {

        pr_debug(".................................START  pkt->rx_src_uport %d\n",pkt->rx_src_uport);
        show_lacpdu(&skb);
        pr_debug(".................................END\n");    

        /* handle the lacp bpdu packet */
        bond_3ad_lacpdu_recv(&skb, pBond, pSlave);
    }

    return SYS_RX_HANDLED;
}

void lacp_task(void *param)
{
    struct bonding *pBond;

    if (TRUE==g_lacp_enable) {

        if (g_u8WalkThroughLagID >= BOARD_MAX_NUM_OF_LAG)
            g_u8WalkThroughLagID = 0;
            
        if (TRUE==g_system_bond.lacp_channel[g_u8WalkThroughLagID].u8enable) {
            pBond = g_system_bond.lacp_channel[g_u8WalkThroughLagID].pBond;

            bond_3ad_state_machine_handler(pBond);
        }
        g_u8WalkThroughLagID++;  
    }
}

void lacp_linkchange(uint16 uport, BOOL link, void *arg)
{
    struct slave *pSlave = NULL;
    struct bonding *pBond = NULL;

    pr_debug("[DBG]: lacp_linkchange(), uport=%d, link=%d\n", uport, link);
   
    /* get the bond/slave pointer by lport */
    if (TRUE==lacp_get_bond_slave_by_lport(uport, NULL, &pBond, &pSlave)) {
        
        bond_update_speed_duplex(uport, pSlave);

        if (TRUE==g_lacp_enable) {

            /* handle the link status/speed/duplex get changed */
            if (TRUE==link)
                bond_3ad_handle_link_change(pSlave, BOND_LINK_UP);
            else 
                bond_3ad_handle_link_change(pSlave, BOND_LINK_DOWN);

            bond_3ad_adapter_speed_changed(pSlave);
            bond_3ad_adapter_duplex_changed(pSlave);
        }
    }
}

void lacp_dbg_print_set(BOOL blflag)
{
    g_lacp_dbg_print = blflag;
    sal_printf("lacp dbug print is %d\n", g_lacp_dbg_print);
}

BOOL lacp_dbg_print_get(void)
{
    return g_lacp_dbg_print;
}


STATICCBK void 
lacp_tx_async_cbk(sys_pkt_t *pkt, sys_error_t status) REENTRANT
{
    if (status != SYS_OK) {
        LACP_ERR(("lacp_tx_async_cbk TX cbk ERROR(%d) !\n", (int16)status)); 
    }
    
    if (pkt) {
        if (pkt->pkt_data) {
            sal_dma_free(pkt->pkt_data);
        }
        sal_free(pkt);
    }
}
/*
 * Hash for the output device based upon layer 2 data
 */
int bond_xmit_hash_policy_l2(struct sk_buff *skb, int count)
{
	struct ethhdr *data = (struct ethhdr *)skb->data;

	return (data->h_dest[5] ^ data->h_source[5]) % count;
}

/**
 * bond_dev_queue_xmit - Prepare skb for xmit.
 *
 * @bond: bond device that got this skb for tx.
 * @skb: hw accel VLAN tagged skb to transmit
 * @slave_dev: slave that is supposed to xmit this skbuff
 */
int bond_dev_queue_xmit(struct bonding *bond, struct sk_buff *skb,
			struct net_device *slave_dev)
{
    sys_error_t retCode;
    uint16 uport;
    BOOL buLink;
    sys_pkt_t *pkt;    

    pr_debug(".................................START\n");
    show_lacpdu(skb);
    pr_debug(".................................END\n");

    if ((NULL==bond) || (NULL==skb) || (NULL==slave_dev))
        return 1;

    uport = slave_dev->uport;
    if (board_port_link_status_get(uport, &buLink) != SYS_OK) {
        return 1;
    }

    if (buLink == FALSE) {
        return 1;
    }

    /* Send lacp packet */
    pkt = (sys_pkt_t *)sal_malloc(sizeof(sys_pkt_t),"_lacp_tx");
    if (pkt == NULL) {
        LACP_ERR(("TX Send Out of memory(pkt)!\n"));
        return 1;
    }   
    sal_memset(pkt, 0, sizeof(sys_pkt_t));
    pkt->pkt_len = skb->len+4;      /* add 4 bytes for CRC */              

    pkt->flags |= SYS_TX_FLAG_USE_UNTAG_PORT_LIST;
    uplist_port_add(pkt->tx_uplist, uport);
    uplist_port_add(pkt->tx_untag_uplist, uport);

    pkt->pkt_data = (uint8 *)sal_dma_malloc(pkt->pkt_len);    
    if (pkt->pkt_data == NULL) {
        sal_free(pkt);
        LACP_ERR(("Out of memory for sending pkt_data (size=%d)!\n", pkt->pkt_len));
        return 1;
    }
    memcpy(pkt->pkt_data, skb->data, skb->len);

    retCode = sys_tx(pkt, lacp_tx_async_cbk);
    if (retCode != SYS_OK) {
        LACP_ERR(("sys_tx() for LACP sending error(%d) !\n", retCode));
    }

	return 0;
}

void lacp_init_cfg(struct lacp_enslave_cfg *placp_cfg)
{
    if (NULL==placp_cfg)
        return;

    memset((char *)placp_cfg, (char)0, sizeof(struct lacp_enslave_cfg));
    //placp_cfg->lacp_activity = LACP_ACTIVE_ACTIVITY;
    placp_cfg->lacp_fast     = LACP_PORT_TIMEOUT;
    placp_cfg->port_priority = LACP_PORT_PRIORITY;
}

/* Return amount of LACP bondings are enabled and calcuate the LAG with active ports. */ 
/*                                                                                    */
/* parameter: The u8LagID is 1 based for LAG ID.                                      */
int lacp_is_enabled(uint8 u8LagID, uint8 *pMemberPorts)
{
    uint8 u8CntLagID, u8Cnt;
    int nActiveLAG = 0, nMemberPorts = 0;
    struct bonding *pBond = NULL;
    struct slave *pSlave = NULL;

    u8LagID = LAGID2TGID(u8LagID);  /* convert to 0 based */    

    if (TRUE==g_lacp_enable) {
        for(u8CntLagID=0; u8CntLagID<BOARD_MAX_NUM_OF_LAG; u8CntLagID++) {
            if (TRUE==g_system_bond.lacp_channel[u8CntLagID].u8enable) {
                pBond = g_system_bond.lacp_channel[u8CntLagID].pBond;
                if (NULL==pBond)
                    continue;

                if (u8LagID==u8CntLagID) {
                    bond_for_each_slave(pBond, pSlave, u8Cnt) {
                        nMemberPorts++;
                    }
                }
                nActiveLAG++;
            }
        }
    }    

    if (NULL!=pMemberPorts)
        *pMemberPorts = (uint8)nMemberPorts;
    return nActiveLAG;
}

/* Check the aggregator is active (running) or not */
BOOL lacp_is_aggregator_active(uint8 u8LagID)
{
    uint8 u8Cnt = 0;
    struct bonding *pBond = NULL;
    struct slave *pSlave = NULL;

    u8LagID = LAGID2TGID(u8LagID);  /* convert to 0 based */    
    if ((FALSE==g_lacp_enable) || (FALSE==g_system_bond.lacp_channel[u8LagID].u8enable))
        return FALSE;

    pBond = g_system_bond.lacp_channel[u8LagID].pBond;
    if (NULL==pBond)
        return FALSE;
        
    bond_for_each_slave(pBond, pSlave, u8Cnt) {
        if (TRUE==pSlave->ad_info.aggregator.is_active)
            return TRUE;
    }
    
    return FALSE;
}

/* Check the LAG & enslaved ports are available for using */ 
int lacp_is_LAG_available(uint8 u8BondMode, struct lacp_enslave_cfg *pLacp_cfg)
{
    uint8 u8Enable, u8LagID, u8CntLagID, u8CurBondMode;
    uint8 u8StaticPorts = 0;
    uint8 u8LacpPorts = 0;    
    uint16 uport, nMaxPorts;
    uint8 lag_uplist[sizeof(pbmp_t)];

    pr_debug("u8BondMode=%d, u8LagID=%d, lport_from=%d, lport_to=%d\n", u8BondMode, pLacp_cfg->u8LagID, pLacp_cfg->uport_from, pLacp_cfg->uport_to);

    u8LagID = LAGID2TGID(pLacp_cfg->u8LagID); /* convert to 0 based */

    /* check the activity mode is the same with the others in LAG */
    if (TRUE==g_system_bond.lacp_channel[u8LagID].u8enable) {
        if (g_system_bond.lacp_channel[u8LagID].blActivity != pLacp_cfg->lacp_activity) {

            LACP_ERR(("The configured LACP activity mode (%d) should be the same in one channel group\n", pLacp_cfg->lacp_activity));
            return SYS_ERR_FALSE;
        }
    }

    nMaxPorts = (uint16)board_uport_count();

    for(u8CntLagID=0; u8CntLagID<BOARD_MAX_NUM_OF_LAG; u8CntLagID++) {

        /* get lacp port bitmap first. since we didnot record the static bond in the lacp's g_system_bond[] */
        memcpy(lag_uplist, g_system_bond.lacp_channel[u8CntLagID].uplist, sizeof(pbmp_t));        

        if (SYS_OK==uplist_is_empty(lag_uplist)) {
            if (SYS_OK != board_lag_group_get(TGID2LAGID(u8CntLagID), &u8Enable, lag_uplist))
                continue;
            if (FALSE==u8Enable)
                continue;

            if (SYS_OK==uplist_is_empty(lag_uplist)) {
                u8CurBondMode = u8BondMode;    
            }
            else
                u8CurBondMode = BOND_MODE_STATIC;
        }
        else
            u8CurBondMode = BOND_MODE_LACP;

        pr_debug("checking u8CntLagID=%d, u8CurBondMode=%d, lag_uplist[]=0x%x 0x%x 0x%x 0x%x\n", u8CntLagID, u8CurBondMode
                    , lag_uplist[0]
                    , lag_uplist[1]
                    , lag_uplist[2]
                    , lag_uplist[3]);

        /* only use the same mode in one LAG */
        if ((u8CurBondMode != u8BondMode) && (u8LagID==u8CntLagID)) {
            if (BOND_MODE_STATIC==u8CurBondMode) {
                LACP_ERR(("The LAG %d is using by Static mode\n", pLacp_cfg->u8LagID));
            }
            else {
                LACP_ERR(("The LAG %d is using by LACP mode\n", pLacp_cfg->u8LagID));
            }
            return SYS_ERR_FALSE;
        }

        /* Check the port was a member of another bond */
        for(uport=1; uport <= nMaxPorts; uport++) {
            if (SYS_OK==uplist_port_matched(lag_uplist, uport)) {

                if (u8LagID==u8CntLagID) {
                    if (BOND_MODE_STATIC==u8CurBondMode)
                        u8StaticPorts++;
                    else
                        u8LacpPorts++;
                }
                else {
                    /* Only using the same mode in one LAG */
                    if ((uport >= pLacp_cfg->uport_from) && (uport <= pLacp_cfg->uport_to)) {
                        if (BOND_MODE_STATIC==u8CurBondMode) {
                            LACP_ERR(("The port %d is a member of Static LAG %d\n", uport+1, u8CntLagID+1));
                        }
                        else {
                            LACP_ERR(("The port %d is a member of LACP LAG %d\n", uport+1, u8CntLagID+1));
                        }
                        return SYS_ERR_EXISTS;
                    }
                }
            }
            else {
                if ((uport >= pLacp_cfg->uport_from) && (uport <= pLacp_cfg->uport_to)) {
                    if (u8LagID==u8CntLagID) {
                        if (BOND_MODE_STATIC==u8CurBondMode)
                            u8StaticPorts++;
                        else
                            u8LacpPorts++;
                    }
                }
            }
        }
    }

    pr_debug("u8StaticPorts=%d, u8LacpPorts=%d\n", u8StaticPorts, u8LacpPorts);

    /* check the new member port + existing member port should not exceed the BOARD_MAX_PORT_PER_LAG */
    if ((u8StaticPorts > BOARD_MAX_PORT_PER_LAG) 
        || (u8LacpPorts > BOARD_MAX_PORT_PER_LAG)) {
        LACP_ERR(("The member ports in LAG %d exceeded the maximum ports %d\n", pLacp_cfg->u8LagID, BOARD_MAX_PORT_PER_LAG));
        return SYS_ERR_FULL;
    }

    return SYS_OK;
}

uint8 lacp_check_lag_type(uint16 uport)
{
    uint8 u8Enable, u8CntLagID;
    uint16 nMaxPorts;
    uint8 lag_uplist[sizeof(pbmp_t)];

    nMaxPorts = (uint16)board_uport_count();

    if ((uport < (uint16)0) || (uport >= nMaxPorts))
        return 0;

    /* Check lacp port bitmap first. since we didnot record the static bond in the lacp's g_system_bond[] */
    for(u8CntLagID=0; u8CntLagID<BOARD_MAX_NUM_OF_LAG; u8CntLagID++) {

        memcpy(lag_uplist, g_system_bond.lacp_channel[u8CntLagID].uplist, sizeof(pbmp_t));        

        if ((SYS_OK==uplist_is_empty(lag_uplist)) || (FALSE==g_system_bond.lacp_channel[u8CntLagID].u8enable)) {
            continue;
        }
       
        if (SYS_OK==uplist_port_matched(lag_uplist, uport))
            return BOND_MODE_LACP;
    }

    /* Check static lag port bitmap. */
    for(u8CntLagID=0; u8CntLagID<BOARD_MAX_NUM_OF_LAG; u8CntLagID++) {
        
        if (SYS_OK==board_lag_group_get(TGID2LAGID(u8CntLagID), &u8Enable, lag_uplist)) 
        {
            if ((SYS_OK==uplist_is_empty(lag_uplist)) || (FALSE==u8Enable)) {
                continue;
            }

            if (SYS_OK==uplist_port_matched(lag_uplist, uport))
                return BOND_MODE_STATIC;
        }
    }

    return 0;
}

/* return the active LAG ID by index (1 based) */
uint8 lacp_get_active_lag(unsigned int nIdx, uint8 *pBondMode)
{
    uint8 u8Enable, u8RetLagId=0;
    uint8 lag_uplist[sizeof(pbmp_t)];

    if (((nIdx < 0) && (nIdx > BOARD_MAX_NUM_OF_LAG)) || (NULL==pBondMode))
        return 0;
    *pBondMode = 0;

    /* get lacp port bitmap fist. since we didnot record the static bond in the lacp's g_system_bond[] */
    memcpy(lag_uplist, g_system_bond.lacp_channel[nIdx-1].uplist, sizeof(pbmp_t));        

    if (SYS_OK==uplist_is_empty(lag_uplist)) {
        if (SYS_OK==board_lag_group_get(nIdx, &u8Enable, lag_uplist)) {
            if (TRUE==u8Enable) {
                *pBondMode = BOND_MODE_STATIC;
                u8RetLagId = nIdx;
            }
        }
    }
    else {
        *pBondMode = BOND_MODE_LACP;
        u8RetLagId = nIdx;        
    }

    pr_debug("nIdx=%d, *pBondMode=%d\n", nIdx, *pBondMode);
    pr_debug("u8RetLagId=%d, get lag_lplist[]=0x%x 0x%x 0x%x 0x%x\n", u8RetLagId
                , lag_uplist[0]
                , lag_uplist[1]
                , lag_uplist[2]
                , lag_uplist[3]);

    return u8RetLagId;  /* return the active LAG by this index */
}

/* Get the member ports by SNMP PortList type and return the bond type.                                             */
/* Each octet within this value specifies a set of eight ports, with the first octet specifying ports 1 through 8,  */
/* the second octet specifying ports 9 through 16, etc. Within each octet, the most significant bit represents      */
/* the lowest numbered port, and the least significant bit represents the highest numbered port.                    */
/* parameters: u8LagID is 1 based.                                                                                  */
uint32 lacp_get_member_ports(uint8 u8LagID, uint8 lag_uplist[sizeof(pbmp_t)])
{ 
    uint8 u8MaxLAGs = board_uport_count()/2;
    uint8 u8Enable;
    uint16 uport, u16Maxport;    
    uint32 u32BondType = 0;
    uint8 temp_uplist[sizeof(pbmp_t)];

    sal_memset(lag_uplist, (char)0, sizeof(pbmp_t));
    sal_memset(temp_uplist, (char)0, sizeof(temp_uplist));    

    if ((u8LagID > 0) && (u8LagID <= u8MaxLAGs))
    {
        if (TRUE==g_system_bond.lacp_channel[u8LagID-1].u8enable)
        {
            /* lacp bond */
            sal_memcpy(temp_uplist, g_system_bond.lacp_channel[u8LagID-1].uplist, sizeof(pbmp_t));

            if (TRUE==g_system_bond.lacp_channel[u8LagID-1].blActivity)
                u32BondType = BOND_MODE_LACP_ACTIVE;
            else
                u32BondType = BOND_MODE_LACP_PASSIVE;
        }
        else
        {
            /* static bond */
            if (SYS_OK==board_lag_group_get(u8LagID, &u8Enable, temp_uplist)) 
            {
                if (TRUE==u8Enable)
                    u32BondType = BOND_MODE_ON;
            }
        }
    }

    /* convert to snmp PortList type */
    u16Maxport = (uint16)board_uport_count();
    for (uport=1; uport<=u16Maxport; uport++) {
        if (SYS_OK==uplist_port_matched(temp_uplist, uport))  {

            lag_uplist[uport / 8] |= (0x80 >> (uport % 8));
        }
    }

    return u32BondType;
}


/* Modify the lplist by the deleting member ports.  */
/* The u8LagID is 1 based.                          */
BOOL lacp_get_port_mapping(uint8 u8BondMode, uint8 u8LagID, uint16 uport_from, uint16 uport_to, uint8 *pCur_uplist, uint8 *pDel_uplist)
{
    uint8 u8Enable;
    uint16 uport, u16Maxport;
    uint8 lag_uplist[sizeof(pbmp_t)];

    if ((NULL==pCur_uplist) || (NULL==pDel_uplist))
        return FALSE;

    u8LagID = LAGID2TGID(u8LagID);  /* convert to 0 based */
    u16Maxport = (uint16)board_uport_count();
    memset(pCur_uplist, (char)0, sizeof(pbmp_t));
    memset(pDel_uplist, (char)0, sizeof(pbmp_t));
    
    if (BOND_MODE_STATIC==u8BondMode) {
        board_lag_group_get(TGID2LAGID(u8LagID), &u8Enable, lag_uplist);
    }
    else if (BOND_MODE_LACP==u8BondMode) {
        memcpy(lag_uplist, g_system_bond.lacp_channel[u8LagID].uplist, sizeof(pbmp_t));
    }
    memcpy(pCur_uplist, lag_uplist, sizeof(pbmp_t));

    for (uport=1; uport<=u16Maxport; uport++) {
        /* update the port bitmap lag_lplist  */    
        if ((SYS_OK==uplist_port_matched(lag_uplist, uport)) && (uport >= uport_from) && (uport <= uport_to)) {
            
            uplist_port_add(pDel_uplist, uport);
        }
    }

    pr_debug("u8LagID=%d, u8BondMode=%d, pCur_uplist[]=0x%x 0x%x 0x%x 0x%x\n"
                , u8LagID
                , u8BondMode
                , pCur_uplist[0]
                , pCur_uplist[1]
                , pCur_uplist[2]
                , pCur_uplist[3]);
    
    pr_debug("u8LagID=%d, u8BondMode=%d, pDel_uplist[]=0x%x 0x%x 0x%x 0x%x\n"
                , u8LagID
                , u8BondMode
                , pDel_uplist[0]
                , pDel_uplist[1]
                , pDel_uplist[2]
                , pDel_uplist[3]);

    return TRUE;
}

/* Get bonding/Slave buffer of LAG ID by logical port */
BOOL lacp_get_bond_slave_by_lport(uint16 uport, uint8 *pGetLagID, struct bonding **pBond, struct slave **pSlave)
{
    BOOL blFoundSlave = FALSE;
    uint8 u8LagID, u8Cnt;
    struct bonding *pCurBond = NULL;
    struct slave *pCurSlave = NULL;

    if (NULL!=pBond)
        *pBond = NULL;
    if (NULL!=pSlave)
        *pSlave = NULL;
    if (NULL!=pGetLagID)
        *pGetLagID = BOARD_MAX_NUM_OF_LAG;
    
    /* get the bond/slave pointer by lport */
    for(u8LagID=0; u8LagID<BOARD_MAX_NUM_OF_LAG; u8LagID++) {
        
        if (SYS_OK == uplist_port_matched(g_system_bond.lacp_channel[u8LagID].uplist, uport)) {
            pCurBond = g_system_bond.lacp_channel[u8LagID].pBond;

            if (NULL==pCurBond)
                break;
            bond_for_each_slave(pCurBond, pCurSlave, u8Cnt) {
                if (pCurSlave->dev->uport==uport) {
                    blFoundSlave = TRUE;

                    if (NULL!=pGetLagID)
                        *pGetLagID = u8LagID+1;
                    if (NULL!=pBond)
                        *pBond = pCurBond;
                    if (NULL!=pSlave)
                        *pSlave = pCurSlave;
                    break;
                }
            }
            break;
        }
    }

    return blFoundSlave;
}

/* Get the full port channel info by LAG ID.                             */
/* Note that the caller should release the returning memory after using. */
/*                                                                       */
/* parameter: u8LagID is 1 based for LAG ID.                             */
BOOL lacp_bond_get_channel_info(uint8 u8LagID, struct port_channel_info *pChannelInfo)
{
    uint8 u8Ports, u8Cnt, u8PortIdx;
    int nActiveLAG;
    rx_states_t sm_rx_state;
    struct bonding *pBond = NULL;
    struct slave *pSlave = NULL;
    struct lacp_port_info *pPortInfo = NULL;

    nActiveLAG = lacp_is_enabled(u8LagID, &u8Ports);

    pr_debug("u8LagID=%d, nActiveLAGs=%d, u8Ports=%d\n", u8LagID, nActiveLAG, u8Ports);

    if ((FALSE==g_lacp_enable) || ((int)0==nActiveLAG) || ((uint8)0==u8Ports) || (NULL==pChannelInfo))
        return FALSE;

    memset((char *)pChannelInfo, (char)0, sizeof(struct port_channel_info));
    u8LagID = LAGID2TGID(u8LagID);  /* convert to 0 based */
    pChannelInfo->u8Ports = u8Ports;
    pChannelInfo->pLacp_port_info = malloc(sizeof(struct lacp_port_info) * u8Ports);
    if (NULL==pChannelInfo->pLacp_port_info)
        return FALSE;
    memset(pChannelInfo->pLacp_port_info, (char)0, sizeof(struct lacp_port_info) * u8Ports);

    pBond = g_system_bond.lacp_channel[u8LagID].pBond;
    u8PortIdx = 0;
    bond_for_each_slave(pBond, pSlave, u8Cnt) {
        
        pPortInfo = &pChannelInfo->pLacp_port_info[u8PortIdx];

        pPortInfo->uport = pSlave->dev->uport;

        /* actor port */
        memcpy(pPortInfo->actor_oper.system.mac_addr_value, SLAVE_AD_INFO(pSlave).port.actor_system.mac_addr_value, sizeof(struct mac_addr));
        pPortInfo->actor_oper.system_priority = SLAVE_AD_INFO(pSlave).port.actor_system_priority;
        pPortInfo->actor_oper.port_number = SLAVE_AD_INFO(pSlave).port.actor_port_number;
        pPortInfo->actor_oper.port_priority = SLAVE_AD_INFO(pSlave).port.actor_port_priority;
        pPortInfo->actor_oper.port_state = SLAVE_AD_INFO(pSlave).port.actor_oper_port_state;
        pPortInfo->actor_oper.key = SLAVE_AD_INFO(pSlave).port.actor_oper_port_key;        

        sm_rx_state = SLAVE_AD_INFO(pSlave).port.sm_rx_state;

        /* partner port */
        memcpy(&pPortInfo->partner_oper, &SLAVE_AD_INFO(pSlave).port.partner_oper, sizeof(struct port_params));

        /* bundled state */
        if ( (((pPortInfo->actor_oper.port_state & AD_STATE_ACTIVE_MODE)==AD_STATE_ACTIVE_MODE) && ((pPortInfo->partner_oper.port_state&AD_STATE_ACTIVE_MODE) == AD_STATE_ACTIVE_MODE))
             || (((pPortInfo->actor_oper.port_state&AD_STATE_ACTIVE_MODE) ==AD_STATE_ACTIVE_MODE) && ((pPortInfo->partner_oper.port_state&AD_STATE_PASSIVE_MODE)==AD_STATE_PASSIVE_MODE))
             || (((pPortInfo->actor_oper.port_state&AD_STATE_PASSIVE_MODE)==AD_STATE_PASSIVE_MODE) && ((pPortInfo->partner_oper.port_state&AD_STATE_ACTIVE_MODE)==AD_STATE_ACTIVE_MODE)))
            strcpy(pPortInfo->szLacpActorState, LACP_STATE_BNDL);

        /* down state */
        else if ((sm_rx_state==AD_RX_LACP_DISABLED) || (sm_rx_state==AD_RX_PORT_DISABLED))
            strcpy(pPortInfo->szLacpActorState, LACP_STATE_DOWN);
        /* hot standby state */
        else if (pPortInfo->actor_oper.port_state & AD_PORT_STANDBY)
            strcpy(pPortInfo->szLacpActorState, LACP_STATE_HOTSBY);    
        /* Independent state */
        else
            strcpy(pPortInfo->szLacpActorState, LACP_STATE_INDEP);

#if 0 /* Ccc: */
        sal_printf("u8LagID=%d, uPorts=%d actor_oper.port_state  %x partner_oper.port_state %x szLacpActorState %s\n", u8LagID, pSlave->dev->uport,pPortInfo->actor_oper.port_state,pPortInfo->partner_oper.port_state,pPortInfo->szLacpActorState );
#endif

        u8PortIdx++;
    }

    return TRUE;
}

/* Add the member ports to enslave device <slave> to bond device <master> */
int lacp_bond_enslave(struct lacp_enslave_cfg *pLacp_cfg)
{
    uint8 u8LagID;
    BOOL buLink;
    sys_error_t nLacpErr = SYS_OK;
    uint16 uport_chk;
    struct net_device *pMasterDev = NULL;
    struct bonding *pBond = NULL;
    struct slave *pNewSlave = NULL;
    struct net_device *pDev = NULL;
    uint8 org_uplist[sizeof(pbmp_t)];     /* original port bitmap */

    pr_debug("pLacp_cfg->u8LagID=%d\n", pLacp_cfg->u8LagID);    
    pr_debug("pLacp_cfg->uport_from=%d\n", pLacp_cfg->uport_from);    
    pr_debug("pLacp_cfg->uport_to=%d\n", pLacp_cfg->uport_to);
    pr_debug("pLacp_cfg->port_priority=%d\n", pLacp_cfg->port_priority);
    pr_debug("pLacp_cfg->lacp_activity=%d\n", pLacp_cfg->lacp_activity);
    pr_debug("pLacp_cfg->lacp_fast=%d\n", pLacp_cfg->lacp_fast);    

    if (NULL==pLacp_cfg)
        return SYS_ERR_PARAMETER;
    if ((pLacp_cfg->u8LagID <= 0) || (pLacp_cfg->u8LagID > BOARD_MAX_NUM_OF_LAG))  
        return SYS_ERR_PARAMETER;
    if (pLacp_cfg->uport_from > pLacp_cfg->uport_to)
        return SYS_ERR_PARAMETER;
    if (pLacp_cfg->port_priority <= 0)
        return SYS_ERR_PARAMETER;
   
    /* check the port is not a member of another LAG and check the LAG is not using by another LAG */
    nLacpErr= lacp_is_LAG_available(BOND_MODE_LACP, pLacp_cfg);
    if (SYS_OK != nLacpErr)
        return nLacpErr;

    /* first time to enslave and should enable lacp and allocate memory pool */
    if (FALSE==g_lacp_enable)
        lacp_enable_set(TRUE);

    u8LagID = LAGID2TGID(pLacp_cfg->u8LagID);   /* convert to 0 based */
    pBond = g_system_bond.lacp_channel[u8LagID].pBond;

    /* allocate memory for bonding if need */
    if (NULL==pBond) {

        pBond = &g_pSysBonds[u8LagID];
        memset(pBond, (char)0, sizeof(struct bonding));
        g_system_bond.lacp_channel[u8LagID].pBond = pBond;

        /* bond transmit policy by layer 2 */
        pBond->xmit_hash_policy = bond_xmit_hash_policy_l2;
        
        /* initiate the aggregation selection timer */
        bond_3ad_initiate_agg_selection(pBond, 1);
    }

    memcpy(org_uplist, g_system_bond.lacp_channel[u8LagID].uplist, sizeof(org_uplist));
    
    /* allocate memory for member ports (slaves) and netdevs */
    for (uport_chk = pLacp_cfg->uport_from;  uport_chk <= pLacp_cfg->uport_to; uport_chk++) {
        
        /* check the device was enslaved and should bypass the process */
        if (SYS_OK == uplist_port_matched(g_system_bond.lacp_channel[u8LagID].uplist, uport_chk)) {
            continue;
        }

        pDev = &g_pSysDevs[uport_chk-1]; //0 base
        pNewSlave = &g_pSysSlaves[uport_chk-1];       
        memset(pDev, (char)0, sizeof(struct net_device));
        memset(pNewSlave, (char)0, sizeof(struct slave));

        /* first slave = master slave */
        if ((u8)0==pBond->slave_cnt) {
            pMasterDev = pDev;    
            pBond->dev = pDev;
            pDev->flags = IFF_MASTER;

            g_system_bond.lacp_channel[u8LagID].blActivity = pLacp_cfg->lacp_activity;
        }
        else {
            pDev->flags = IFF_SLAVE;
            pMasterDev = pBond->dev->master;
        }
    
        /* assign the initial value for the net device */
        pDev->uport = uport_chk;
        pDev->bond = pBond;
        pDev->master = pMasterDev;
        sprintf(pDev->name, "%s%d", ETH_NAME, pDev->uport);
        
        get_system_mac(pDev->dev_addr);

        /* assign the initial value for the slave */
        pNewSlave->dev = pDev;
        pNewSlave->bond = pBond;

        /* in 802.3ad mode, the internal mechanism will activate 
         * the slaves in the selected aggregator.
         */
        bond_set_slave_inactive_flags(pNewSlave);

        /* get the link status, speed and duplex */
        bond_update_speed_duplex(pDev->uport, pNewSlave);

        /*
         * Save slave's original ("permanent") mac address for modes
         * that need it, and for restoring it upon release, and then
         * set it to the master's address
         */
        memcpy(pNewSlave->perm_hwaddr, pDev->dev_addr, ETH_ALEN);
        //memcpy(pDev->dev_addr, pMasterDev->dev_addr, ETH_ALEN);
        
        bond_attach_slave(pBond, pNewSlave);
    
        /* Initialize AD with the number of times that the AD timer is called in 1 second
         * can be called only after the mac address of the bond is set
         */
        if ((u8)1==pBond->slave_cnt) {
            bond_3ad_initialize(pBond, (u16)sal_get_us_per_tick(), g_system_bond.lacp_sys_info.sys_priority);
        }

        //board_to_user_port(pDev->lport, &uport);
        SLAVE_AD_INFO(pNewSlave).id = pDev->uport;//;uport+1;
        bond_3ad_bind_slave(pNewSlave, pLacp_cfg->lacp_fast, pLacp_cfg->lacp_activity, pLacp_cfg->port_priority);

        /* update the port bitmap */
        uplist_port_add(g_system_bond.lacp_channel[u8LagID].uplist, uport_chk);
    }
    
    lacp_update_dev_flag(u8LagID);

    /* Set the LAG to HW */
    g_system_bond.lacp_channel[u8LagID].u8enable = TRUE;    
    nLacpErr = board_lag_group_set(pLacp_cfg->u8LagID, TRUE, g_system_bond.lacp_channel[u8LagID].uplist, (uint8)0);

    pr_debug("board_lag_group_set(), u8LagID=%d, uplist[]=0x%x,0x%x,0x%x,0x%x, errcode=%d !\n"
                , pLacp_cfg->u8LagID
                , g_system_bond.lacp_channel[u8LagID].uplist[0]
                , g_system_bond.lacp_channel[u8LagID].uplist[1]
                , g_system_bond.lacp_channel[u8LagID].uplist[2]
                , g_system_bond.lacp_channel[u8LagID].uplist[3]
                , nLacpErr);

    if (SYS_OK==nLacpErr) {
        bond_3ad_set_carrier(pBond);

        for (uport_chk = pLacp_cfg->uport_from;  uport_chk <= pLacp_cfg->uport_to; uport_chk++) {
            /* The existed/running LACP port should not make it to re-sync */
            if ((FALSE==pLacp_cfg->cfgInitial) && (SYS_OK == uplist_port_matched(org_uplist, uport_chk))) {
                continue;
            }

            /* link down then link up the port to cause the link change event when the port was link up */
            if (SYS_OK==board_port_link_status_get(uport_chk, &buLink)) {
                if (buLink) {
                    board_port_enable_set(uport_chk, (uint8)0);
                    sal_sleep(300);
                    board_port_enable_set(uport_chk, (uint8)1);
                }
            }
        }
    }
    else
    {
        lacp_bond_release(TRUE, pLacp_cfg->u8LagID, pLacp_cfg->uport_from, pLacp_cfg->uport_to);
        LACP_ERR(("board_lag_group_set() fail, u8LagID=%d, from port %d to port %d, errcode=%d !\n", pLacp_cfg->u8LagID, pLacp_cfg->uport_from+1, pLacp_cfg->uport_to+1, nLacpErr));
    }

    return nLacpErr;
}

/*
 * Delete the member ports to release specified slaves in the bond. 
 * 
 * parameters: u8LagID is 1 based for LAG ID.
 */
int lacp_bond_release(BOOL blDefaulted, uint8 u8LagID, uint16 uport_from, uint16 uport_to)
{
    uint8 u8Enable;
    sys_error_t nLacpErr = SYS_OK;
    int nCnt, nCntMaxSlave;
    uint16 uport, maxPorts;
    struct bonding *pBond = NULL;
    struct slave *pSlave = NULL;
    struct slave *pNextSlave;
    uint8 cur_uplist[sizeof(pbmp_t)];
    uint8 del_uplist[sizeof(pbmp_t)];    
    uint8 unit,lport=0;
    
    pr_debug("u8LagID=%d, uport_from=%d, uport_to=%d\n", u8LagID, uport_from, uport_to);

    if ((u8LagID <= 0) || (u8LagID > BOARD_MAX_NUM_OF_LAG))  
        return SYS_ERR_PARAMETER;
    if (uport_from > uport_to)
        return SYS_ERR_PARAMETER;
    if ((FALSE==g_system_bond.lacp_channel[u8LagID-1].u8enable) && (FALSE==blDefaulted)){
        return SYS_ERR_PARAMETER;
    }        
    if (NULL==g_system_bond.lacp_channel[u8LagID-1].pBond){
        return SYS_ERR_PARAMETER;
    }
    if (FALSE==lacp_get_port_mapping(BOND_MODE_LACP, u8LagID, uport_from, uport_to, cur_uplist, del_uplist)){
        return SYS_ERR_PARAMETER;
    }        
    if (SYS_OK==uplist_is_empty(del_uplist)){
        return SYS_ERR_PARAMETER;
    }
    u8LagID = LAGID2TGID(u8LagID);   /* convert to 0 based */
    pBond = g_system_bond.lacp_channel[u8LagID].pBond;

    /* update the port bitmap */
    if (!memcmp(cur_uplist, del_uplist, sizeof(pbmp_t))) {
        u8Enable = 0;   /* remove all member ports and its LAG */
        g_system_bond.lacp_channel[u8LagID].blActivity = LACP_PASSIVE_ACTIVITY;
        memset(g_system_bond.lacp_channel[u8LagID].uplist, (char)0, sizeof(pbmp_t));
    }
    else {
        u8Enable = 1;   /* some of member ports are removed and changed to modify */
        for (nCnt=0; nCnt<sizeof(pbmp_t); nCnt++)
            g_system_bond.lacp_channel[u8LagID].uplist[nCnt] ^= del_uplist[nCnt];
    }

    /*
    pr_debug("new g_system_bond.lacp_channel[%d].lplist=0x%x,0x%x,0x%x,0x%x \n"
                , u8LagID
                , g_system_bond.lacp_channel[u8LagID].lplist[0]
                , g_system_bond.lacp_channel[u8LagID].lplist[1]
                , g_system_bond.lacp_channel[u8LagID].lplist[2]
                , g_system_bond.lacp_channel[u8LagID].lplist[3]);
    */

    /* release the specified slave/dev in this bond */
    nCntMaxSlave = pBond->slave_cnt;
    for (nCnt=0, pSlave=pBond->first_slave; nCnt<nCntMaxSlave; nCnt++) {
        pNextSlave = pSlave->next;
        if (SYS_OK==uplist_port_matched(del_uplist, pSlave->dev->uport))
        {
            bond_3ad_unbind_slave(pSlave);
            bond_detach_slave(pBond, pSlave);

            /* ignore to restore original ("permanent") mac address, set new master dev and dev_addr */
        }
        pSlave = pNextSlave;        
    }

    /* change the port state back to forwarding */
    maxPorts = (uint16)board_uport_count();
    for(uport=1; uport<=maxPorts; uport++) {
        if (uplist_port_matched(del_uplist, uport) == SYS_OK) {
            board_uport_to_lport(uport, &unit, &lport);
            //stp_portState_set(uport+1, LACP_PORTSTATE, STP_FORWARDING);
            board_stp_port_set(lport, STP_FORWARDING);
        }
    }   

    /* Modify the LAG to HW */
    if (u8Enable)
        memcpy(del_uplist, g_system_bond.lacp_channel[u8LagID].uplist, sizeof(pbmp_t));
        
    if (TRUE==blDefaulted)
        memset(del_uplist, (char)0, sizeof(del_uplist));

    nLacpErr = board_lag_group_set(TGID2LAGID(u8LagID), u8Enable, del_uplist, (uint8)0);
    if (SYS_OK != nLacpErr) {
        LACP_ERR(("board_lag_group_set(), u8LagID=%d, u8Enable=%d, del_lplist[]=0x%x,0x%x,0x%x,0x%x, errcode=%d !\n"
                    , u8LagID
                    , u8Enable
                    , del_uplist[0]
                    , del_uplist[1]
                    , del_uplist[2]
                    , del_uplist[3]
                    , nLacpErr));
    }    
    
    if (NULL!=pBond) {
        if ((u8)0==pBond->slave_cnt) {
            g_system_bond.lacp_channel[u8LagID].pBond = NULL;
            g_system_bond.lacp_channel[u8LagID].u8enable = FALSE;

            /* not found any active lacp and should disable lacp and free the memory pool */
            if ((int)0==lacp_is_enabled(u8LagID, NULL))
                lacp_enable_set(FALSE);
         }
    }

    lacp_update_dev_flag(u8LagID);
  
    return nLacpErr;
}

/*
 * Delete the channel: Try to release all slaves in the bond
 *
 * parameter: The u8LagID is 1 based for LAG ID.
 */
int lacp_bond_release_all(BOOL blDefaulted, uint8 u8LagID)
{
    sys_error_t nLacpErr = SYS_OK;
    uint16 uport, uport_from, uport_to, maxPorts;

    if ((u8LagID <= 0) || (u8LagID > BOARD_MAX_NUM_OF_LAG))  
        return SYS_ERR_PARAMETER;

    maxPorts = (uint16)board_uport_count()+1; //avoid no port member in trunk group

    for(uport=1, uport_from=maxPorts, uport_to=maxPorts; uport<=maxPorts; uport++) {
        if (uplist_port_matched(g_system_bond.lacp_channel[u8LagID-1].uplist, uport) == SYS_OK) {

            if (maxPorts==uport_from) {
                uport_from = uport;
                uport_to = uport;
            }
            if (uport > uport_to)
                uport_to = uport;
        }
    }
    if ((maxPorts != uport_from) && (maxPorts != uport_to)) {//avoid no port member in trunk group
        nLacpErr = lacp_bond_release(blDefaulted, u8LagID, uport_from, uport_to);
        if ((SYS_OK==nLacpErr) || (TRUE==blDefaulted))
            memset(&g_system_bond.lacp_channel[u8LagID-1], (char)0, sizeof(struct bond_node));
    }
    return nLacpErr;
}

void lacp_set_sys_priority(uint16 sys_priority)
{
    if (sys_priority <= 0)
        sys_priority = DEFAULT_LACP_SYS_PRIORITY;

    g_system_bond.lacp_sys_info.sys_priority = sys_priority;
}

int lacp_set_port_priority(uint16 uport, uint16 port_priority)
{
    if (SAL_UPORT_IS_NOT_VALID(uport))
        return SYS_ERR_PARAMETER;

    if (port_priority <= 0)
        port_priority = DEFAULT_LACP_PORT_PRIORITY;


    g_system_bond.lacp_sys_info.sys_port_config[SAL_UPORT_TO_ZUPORT(uport)].port_priority = port_priority;
    return SYS_OK;
}
void lacp_get_bond_info(struct lacp_bond *pLacpBondInfo)
{
    if (NULL!=pLacpBondInfo)
        memcpy(pLacpBondInfo, &g_system_bond, sizeof(struct lacp_bond));
}

BOOL lacp_enable_set(uint8 enable)
{
    BOOL blResult = TRUE;
    uint8 u8CntLagID;

    if (TRUE==enable) {
        /* enabling LACP */
        if (FALSE==g_lacp_enable) {

            if(sys_rx_register(lacp_rx, LACP_PRIO, NULL, 0) != SYS_OK) {
                LACP_ERR(("sys_rx_register failed in lacp_enable_set()\n"));
                blResult = FALSE;
            }

            if (FALSE==sys_register_linkchange(lacp_linkchange, NULL))
                blResult = FALSE;

            /* allocate big memory pool for all bonds/slaves/devs */
            blResult = allocateLacpMemoryPool();

            if (TRUE==blResult) {
                g_lacp_enable= TRUE;
                
                /* enable LACP BPDU to L2_USER_ENTRY */
                board_bpdu_enable_set(LACP_BPDU_INDEX, enable);
            }
         }
    }

    /* disabling LACP */
    if ((FALSE==enable) || (FALSE==blResult))
    {
        g_lacp_enable = FALSE;

        /* disable LACP BPDU to L2_USER_ENTRY */
        board_bpdu_enable_set(LACP_BPDU_INDEX, enable);
        
        sys_rx_unregister(lacp_rx);
        sys_unregister_linkchange(lacp_linkchange);

        /* release all bonds */
        for(u8CntLagID=1; u8CntLagID<=BOARD_MAX_NUM_OF_LAG; u8CntLagID++) {
            
            if (SYS_OK != lacp_bond_release_all(TRUE, u8CntLagID)) {
                blResult = FALSE;
                LACP_ERR(("release LAG %d is failed\n", u8CntLagID));
            }
        }

        /* free lacp big memory pool for all bonds/slaves/devs */
        if (TRUE==blResult)
            freeLacpMemoryPool();
    }

    pr_debug("enable=%d, g_lacp_enable=%d, blResult=%d\n", enable, g_lacp_enable, blResult);    
    return blResult;
}

void lacp_init(void)
{
    memset((char *)&g_system_bond, (char)0, sizeof(g_system_bond));

    /* Only a single combination of Actor System ID and System Priority are considered, and no distinction is 
       made between the values of these parameters for an Aggregator and the port(s) that are associated with it. */
    lacp_set_sys_priority(DEFAULT_LACP_SYS_PRIORITY);   /* set the default system priority */
    get_system_mac(g_system_bond.lacp_sys_info.sys_mac_addr.mac_addr_value);
    
    task_add(lacp_task, (void *)NULL);   
}

#endif /* CFG_SWITCH_LAG_INCLUDED */
